LICENSE_TEXT = """Epoptes is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or any later version.

Epoptes is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Epoptes. If not, see <http://www.gnu.org/licenses/>."""

__version__ = "1.0"  # Automatically updated while packaging
__license__ = "GPL"
